"""
    format_type(Vk.FORMAT_R4G4_UNORM_PACK8) # UInt8
    format_type(Vk.FORMAT_R32_SFLOAT) # Float32
    format_type(Vk.FORMAT_R32G32_SFLOAT) # NTuple{2,Float32}

    format_type(Vk.FORMAT_R32G32B32_SFLOAT) # RGB{Float32} with the extension for ColorTypes.jl
    format_type(Vk.FORMAT_R16G16B16A16_SFLOAT) # RGBA{Float16} with the extension for ColorTypes.jl

Retrieve a canonical type associated with an uncompressed Vulkan image format.

Note from the spec:
> Depth/stencil formats are considered opaque and need not be stored in the exact number of bits per texel or component ordering indicated by the format enum.

This means we can't reliably interpret an image with a depth/stencil format. The image needs to be transitioned
to a color format first (e.g. `D16` to `R16`), and when both depth and stencil aspects are present,
a view must be formed for the transfer (e.g. `D16S8` must be viewed with the depth aspect for transfer from `D16` into `R16`, or with the stencil aspect for transfer from `S8` into `R8`).

The exact byte representation is available at https://registry.khronos.org/vulkan/specs/1.3-extensions/html/chap49.html#texel-block-size.

Note from the spec for packed representations:
> Packed formats store multiple components within one underlying type. The bit representation is that the first component specified in the name of the format is in the most-significant bits and the last component specified is in the least-significant bits of the underlying type. The in-memory ordering of bytes comprising the underlying type is determined by the host endianness.

One must therefore be careful about endianness for packed representations when reading from an image.

# Extended help

Here is an informative list of most mappings (the element type, where relevant, is omitted and represented as `T`):

`PACK8`: -> `UInt8`
- `RG`

`PACK16`: -> `UInt16`
- `RGBA`
- `BGRA`
- `RGB`
- `BGR`
- `RGBA1`
- `BGRA1`
- `A1RGB`
- `A4RGB`
- `A4BGR`
- `R12X4`

`PACK32`: -> `UInt32`
- `ARGB`
- `A2RGB`
- `A2BGR`
- `BGR`
- `EBGR`
- `X8D24`
- `GBGR_422`
- `BGRG_422`

8-bit per component:
- `R` -> `T`
- `RG` -> `NTuple{2,T}`
- `RGB` -> `RGB{T}`
- `BGR` -> `BGR{T}`
- `RGBA` -> `RGBA{T}`
- `BGRA` -> `BGRA{T}`
- `ABGR` -> `ABGR{T}`
- `GBGR` -> `NTuple{4,T}`
- `BGRG` -> `NTuple{4,T}`
- `S` -> undefined, transition to `R8`

16-bit per component:
- `R` -> `T`
- `RG` -> `NTuple{2,T}`
- `RGB` -> `RGB{T}`
- `RGBA` -> `RGBA{T}`
- `D` -> undefined, transition to `R16`

32-bit per component:
- `R` -> `T`
- `RG` -> `NTuple{2,T}`
- `RGB` -> `RGB{T}`
- `RGBA` -> `RGBA{T}`
- `D` -> undefined, transition to `R32`

64-bit per component:
- `R` -> `T`
- `RG` -> `NTuple{2,T}`
- `RGB` -> `RGB{T}`
- `RGBA` -> `RGBA{T}`

Depth/stencil:
- `D16S8` -> undefined, transition to `R16`/`R8`
- `D24S8` -> undefined, transition to ?/`R8`
- `D32S8` -> undefined, transition to `R32`/`R8`

Compressed formats: -> undefined byte representation, transition to other format
- `BC`
- `ETC2`
- `EAC`
- `ASTC`
- `PVRTC`
"""
function format_type end

@inline format_type(x) = format_type(Val(x))
format_type(::Val{x}) where {x} = error("No type is known which corresponds to format $x")

# generated by `ext/generate_formats.jl`.

Vk.format_type(::Val{Vk.FORMAT_R4G4_UNORM_PACK8}) = UInt8
Vk.format_type(::Val{Vk.FORMAT_R4G4B4A4_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_B4G4R4A4_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R5G6B5_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_B5G6R5_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R5G5B5A1_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_B5G5R5A1_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_A1R5G5B5_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_UNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_SNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_USCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_SSCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_UINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_SINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A8B8G8R8_SRGB_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_UNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_SNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_USCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_SSCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_UINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2R10G10B10_SINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_UNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_SNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_USCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_SSCALED_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_UINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_A2B10G10R10_SINT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_B10G11R11_UFLOAT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_E5B9G9R9_UFLOAT_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_X8_D24_UNORM_PACK32}) = UInt32
Vk.format_type(::Val{Vk.FORMAT_R10X6_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R10X6G10X6_UNORM_2PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R10X6G10X6B10X6A10X6_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6B10X6G10X6R10X6_422_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_B10X6G10X6R10X6G10X6_422_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6_R10X6_3PLANE_420_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6R10X6_2PLANE_420_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6_R10X6_3PLANE_422_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6R10X6_2PLANE_422_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6_R10X6_3PLANE_444_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R12X4_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R12X4G12X4_UNORM_2PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_R12X4G12X4B12X4A12X4_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4B12X4G12X4R12X4_422_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_B12X4G12X4R12X4G12X4_422_UNORM_4PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4_R12X4_3PLANE_420_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4R12X4_2PLANE_420_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4_R12X4_3PLANE_422_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4R12X4_2PLANE_422_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4_R12X4_3PLANE_444_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_A4R4G4B4_UNORM_PACK16}) = UInt16
Vk.format_type(::Val{Vk.FORMAT_A4B4G4R4_UNORM_PACK16}) = UInt16

Vk.Format(::Type{N0f8}) = FORMAT_R8_UNORM
Vk.Format(::Type{Q0f7}) = FORMAT_R8_SNORM
Vk.Format(::Type{Tuple{N0f8, N0f8}}) = FORMAT_R8G8_UNORM
Vk.Format(::Type{Tuple{Q0f7, Q0f7}}) = FORMAT_R8G8_SNORM
Vk.Format(::Type{N0f16}) = FORMAT_R16_UNORM
Vk.Format(::Type{Q0f15}) = FORMAT_R16_SNORM
Vk.Format(::Type{Float16}) = FORMAT_R16_SFLOAT
Vk.Format(::Type{Tuple{N0f16, N0f16}}) = FORMAT_R16G16_UNORM
Vk.Format(::Type{Tuple{Q0f15, Q0f15}}) = FORMAT_R16G16_SNORM
Vk.Format(::Type{Tuple{Float16, Float16}}) = FORMAT_R16G16_SFLOAT
Vk.Format(::Type{Float32}) = FORMAT_R32_SFLOAT
Vk.Format(::Type{Tuple{Float32, Float32}}) = FORMAT_R32G32_SFLOAT
Vk.Format(::Type{Float64}) = FORMAT_R64_SFLOAT
Vk.Format(::Type{Tuple{Float64, Float64}}) = FORMAT_R64G64_SFLOAT

Vk.format_type(::Val{FORMAT_R8_UNORM}) = N0f8
Vk.format_type(::Val{FORMAT_R8_SNORM}) = Q0f7
Vk.format_type(::Val{FORMAT_R8G8_UNORM}) = Tuple{N0f8, N0f8}
Vk.format_type(::Val{FORMAT_R8G8_SNORM}) = Tuple{Q0f7, Q0f7}
Vk.format_type(::Val{FORMAT_R16_UNORM}) = N0f16
Vk.format_type(::Val{FORMAT_R16_SNORM}) = Q0f15
Vk.format_type(::Val{FORMAT_R16_SFLOAT}) = Float16
Vk.format_type(::Val{FORMAT_R16G16_UNORM}) = Tuple{N0f16, N0f16}
Vk.format_type(::Val{FORMAT_R16G16_SNORM}) = Tuple{Q0f15, Q0f15}
Vk.format_type(::Val{FORMAT_R16G16_SFLOAT}) = Tuple{Float16, Float16}
Vk.format_type(::Val{FORMAT_R32_SFLOAT}) = Float32
Vk.format_type(::Val{FORMAT_R32G32_SFLOAT}) = Tuple{Float32, Float32}
Vk.format_type(::Val{FORMAT_R64_SFLOAT}) = Float64
Vk.format_type(::Val{FORMAT_R64G64_SFLOAT}) = Tuple{Float64, Float64}
Vk.format_type(::Val{FORMAT_G8B8G8R8_422_UNORM}) = NTuple{4, N0f8}
Vk.format_type(::Val{FORMAT_B8G8R8G8_422_UNORM}) = NTuple{4, N0f8}
Vk.format_type(::Val{FORMAT_G16B16G16R16_422_UNORM}) = NTuple{4, N0f16}
Vk.format_type(::Val{FORMAT_B16G16R16G16_422_UNORM}) = NTuple{4, N0f16}
